// VDP.CPP
// Win32 VDP1/VDP2 Drawing Routines
// (should eventually port to plugin for Win32 version)

#include <windows.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ddraw.h>
#include "vdp.h"

HWND vdpWnd;
unsigned char *fb, *vram;
unsigned short  *vdp1reg, *vdp2col, *vdp2reg;

int screenMode, sizex=320, sizey=240, scrnx, scrny, depth;
unsigned char cram;

int initScreenWindowed(void);
int initFulScreen(void);

static LPDIRECTDRAW			lpDD;
static LPDIRECTDRAWSURFACE	lpDDSPrimary;
static LPDIRECTDRAWSURFACE	lpDDSBack;
static LPDIRECTDRAWSURFACE	lpDDSBkgrnd;
static LPDIRECTDRAWCLIPPER	pcClipper;
static DDSURFACEDESC		ddsd;
static DDSCAPS				ddscaps;
static HRESULT				ddrval;

static DDBLTFX     ddbltfx;

vdp_pal palette[0x1000];

/*tagPlugInfo info;
tagPlugInfo PLUGINCALL SAT_PlugInfo(void)
{
	strcpy(info.name, "VDP DirectX Plugin");
	info.plugType = SAT_VDP_PLUGIN;
	info.plugVersion = 1;

	return info;
}*/

int PLUGINCALL VDP_InitPlugin(unsigned char *VDP1, unsigned char *VDP2, unsigned long param)
{
	//HKEY myKey;
	fb      = VDP1;
	vdp1reg = (unsigned short*)(VDP1 + 0x30000);
	vram    = VDP2;
	vdp2col = (unsigned short*)(VDP2 + 0x80000);
	vdp2reg = (unsigned short*)(VDP2 + 0x81000);
	
	// param contains hWnd
	vdpWnd = (HWND)param;

	if (DirectDrawCreate(NULL, &lpDD, NULL) != DD_OK) 
		return -1;

	lpDD->Release();
	lpDD = NULL;

	return 0;
}

int PLUGINCALL VDP_StartPlugin(void)
{
	if(DirectDrawCreate(NULL, &lpDD, NULL) != DD_OK )
		return -1;

	return initScreenWindowed();
}

void PLUGINCALL VDP_StopPlugin(void)
{
	if(lpDD != NULL)
    {
		if (screenMode | VDP_FULLSCREEN) lpDD->RestoreDisplayMode();

        if(lpDDSBack != NULL)
        {
            lpDDSBack->Release();
            lpDDSBack = NULL;
        }
        if(lpDDSPrimary != NULL)
        {
            lpDDSPrimary->Release();
            lpDDSPrimary = NULL;
        }
		lpDD->Release();
		lpDD = NULL;
    }
}

void PLUGINCALL VDP_FreePlugin(void)
{}

void PLUGINCALL VDP_UpdateScreen(void)
{
}

extern void ConsoleMsg(int, char*, ...);

void PLUGINCALL VDP_UpdateRegs(unsigned long addr)
{
	int i, tmp;
	char tmpchar;

	switch (addr & 0xffffff)
	{
		case (0xf8000e):
			cram = (VDP2_RAMCTL>>12)&0x3;
			switch(cram) {
				case 0: //CRAM mode 0: RGB each 5 bits, 1024 color setting
					for(i=0; i<0x10; i++)
					{
						tmp = *(vdp2col+i) & 0x7fff;
						palette[i].red   = (tmp & 0x1f) << 3;
						palette[i].green = ((tmp >> 5) & 0x1f) << 3;
						palette[i].blue  = (tmp >> 10) << 3;
					} break;

				//case 1: //CRAM mode 1: RGB each 5 bits, 2048 color setting
				//case 2: //CRAM mode 2: RGB each 8 bits, 1024 color setting

				default: ConsoleMsg(0, "Unhandled CRAM mode %d", cram);
			} break;

		default: ConsoleMsg(0, "Unhandled VDP Register: 0x%8x", addr); break;
	}
}

void PLUGINCALL VDP_VBlank(void)
{}

void PLUGINCALL VDP_HBlank(void)
{}

int initFullScreen(void)
{
    if (lpDD->SetCooperativeLevel(vdpWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN) != DD_OK)
		return -1;

	// should really enumerate x, y, and depth before this
	if (lpDD->SetDisplayMode(sizex, sizey, depth) != DD_OK)
		return -1;

    // Create the primary surface
	memset(&ddsd, 0, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;

    if (lpDD->CreateSurface(&ddsd, &lpDDSPrimary, NULL) != DD_OK)
		return -1;

	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;

	if (lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSBack) != DD_OK)
		return -1;

	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0;
   
    lpDDSPrimary->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
    lpDDSBack->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
	screenMode |= VDP_FULLSCREEN;

	return 0;
}

int initScreenWindowed(void)
{
	RECT rcSize;


    if(lpDD->SetCooperativeLevel(vdpWnd, DDSCL_NORMAL) != DD_OK)
		return -1;

	// Resize window
	GetClientRect(vdpWnd, &rcSize); 
	rcSize.right = rcSize.left + sizex;
	rcSize.bottom = rcSize.top + sizey;
	AdjustWindowRect(&rcSize, WS_OVERLAPPED | WS_SYSMENU | WS_BORDER, TRUE);

    // Create the primary surface
	memset(&ddsd, 0, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

    if (lpDD->CreateSurface(&ddsd, &lpDDSPrimary, NULL) != DD_OK)
		return -1;

	// Create off-screen surface
    ddsd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
    ddsd.dwWidth  = rcSize.right - rcSize.left;
    ddsd.dwHeight = rcSize.bottom - rcSize.top;
 
	if (lpDD->CreateSurface(&ddsd, &lpDDSBack, NULL) != DD_OK )
		return -1;

	if (lpDD->CreateClipper(NULL, &pcClipper, NULL) != DD_OK)
		return -1;
	
    pcClipper->SetHWnd(NULL, vdpWnd);
    lpDDSPrimary->SetClipper(pcClipper);
    pcClipper->Release();
	screenMode |= VDP_WINDOWED;

	return 0;
}