// Global.h
// Global Settings

#ifndef _GLOBAL_H
#define _GLOBAL_H

#include <stdarg.h>

#define uint8	unsigned char
#define uint16	unsigned short
#define uint32	unsigned long
#ifdef __GNUC__
#define uint64	unsigned long int
#define sint64	long int

#define FALSE	0
#define TRUE	1
#else
#define uint64	unsigned __int64
#define sint64	__int64
#endif

#define MSG_NORM	0
#define MSG_WARN	1
#define MSG_FATAL	2

#define EMU_STOP	0
#define EMU_RUN		1
#define EMU_PAUSE	2

// Global Options
//#define CPU_PROFILING
//#define TITAN_DEBUG

#define swapWord(x)	(uint16)(((x)>>8)|((x)<<8))
extern inline unsigned long swapLong(unsigned long);

#ifdef WIN32
#include <windows.h>

extern HWND g_hWnd;
extern HINSTANCE g_hInst;
#endif

#ifdef TITAN_DEBUG
#include "debug.h"

#define MSG_DEBUG	3
#endif

extern uint8 EmuStatus;

typedef struct {
	uint8	CountryCode;
	uint8	ConsoleLog;
	uint8	Engine;
	uint8	AreaCode;
	char VDPModule[255], SPUModule[255], SCUModule[255], PIOModule[255];
} tagPrefs;

extern tagPrefs Settings;

typedef struct {
	uint8	EmuStatus;
	uint8	SlaveEnabled;
	uint8	SoundEnabled;
	uint8	ResetEnabled;
} tagInternal;

extern tagInternal Flags;

// Console & Settings Functions
int emuStartUp(void);
void emuCleanUp(void);
int loadINISettings(void);
void ConsoleMsg(int flag, char *message, ...);

// Emulation Functions
void ResetCPU(void);
void emulate(void);
void dynaEmulate(void);

//Plugin Functions
extern int initModules(void);
extern int startModules(void);
extern void stopModules(void);
extern void freeModules(void);

#endif