// SH2.CPP
// SH2 CPU Emulation Routines

#include <stdlib.h>
#include <string.h>
#include "global.h"
#include "mem.h"
#include "sh2.h"

#define source		(uint8)((opcode >> 4) & 0xf) //m
#define dest		(uint8)((opcode >> 8) & 0xf) //n
#define offset1		(opcode & 0xf)
#define offset2		(opcode & 0xff)
#define offset3		(opcode & 0xfff)
#define immediate	(signed char)offset2

#define setTBit(x)		SR = (SR & 0xfffffffe) | x

#define setMBit(x)		SR = (SR & 0xfffffdff) | ((x&1) << 9)
#define setQBit(x)		SR = (SR & 0xfffffeff) | ((x&1) << 8)
#define setSBit(x)		SR = (SR & 0xfffffffd) | ((x&1) << 1)

#define DelayBranch(x)	PC += 2; \
						DelayWait = x; \
						doInst(sh2, memGetWord(PC)); \
						PC = DelayWait - 2; //compensate for "stupid" emulate() func

tagSH2 master, slave;

i_descr opcode_list[] =
{
        { ZERO_F,	"clrt",                0xffff, 0x8,    0,      0 },
        { ZERO_F,   "clrmac",              0xffff, 0x28,   0,      0 },
        { ZERO_F,   "div0u",               0xffff, 0x19,   0,      0 },
        { ZERO_F,   "nop",                 0xffff, 0x9,    0,      0 },
        { ZERO_F,   "rte",                 0xffff, 0x2b,   0,      0 },
        { ZERO_F,   "rts",                 0xffff, 0xb,    0,      0 },
        { ZERO_F,   "sett",                0xffff, 0x18,   0,      0 },
        { ZERO_F,   "sleep",               0xffff, 0x1b,   0,      0 },
        { N_F,      "cmp/pl r%d",          0xf0ff, 0x4015, 0,      0 },
        { N_F,      "cmp/pz r%d",          0xf0ff, 0x4011, 0,      0 },
        { N_F,      "dt r%d",              0xf0ff, 0x4010, 0,      0 },
        { N_F,      "movt r%d",            0xf0ff, 0x0029, 0,      0 },
        { N_F,      "rotl r%d",            0xf0ff, 0x4004, 0,      0 },
        { N_F,      "rotr r%d",            0xf0ff, 0x4005, 0,      0 },
        { N_F,      "rotcl r%d",           0xf0ff, 0x4024, 0,      0 },
        { N_F,      "rotcr r%d",           0xf0ff, 0x4025, 0,      0 },
        { N_F,      "shal r%d",            0xf0ff, 0x4020, 0,      0 },
        { N_F,      "shar r%d",            0xf0ff, 0x4021, 0,      0 },
        { N_F,      "shll r%d",            0xf0ff, 0x4000, 0,      0 },
        { N_F,      "shlr r%d",            0xf0ff, 0x4001, 0,      0 },
        { N_F,      "shll2 r%d",           0xf0ff, 0x4008, 0,      0 },
        { N_F,      "shlr2 r%d",           0xf0ff, 0x4009, 0,      0 },
        { N_F,      "shll8 r%d",           0xf0ff, 0x4018, 0,      0 },
        { N_F,      "shlr8 r%d",           0xf0ff, 0x4019, 0,      0 },
        { N_F,      "shll16 r%d",          0xf0ff, 0x4028, 0,      0 },
        { N_F,      "shlr16 r%d",          0xf0ff, 0x4029, 0,      0 },
        { N_F,      "stc sr, r%d",         0xf0ff, 0x0002, 0,      0 },
        { N_F,      "stc gbr, r%d",        0xf0ff, 0x0012, 0,      0 },
        { N_F,      "stc vbr, r%d",        0xf0ff, 0x0022, 0,      0 },
        { N_F,      "sts mach, r%d",       0xf0ff, 0x000a, 0,      0 },
        { N_F,      "sts macl, r%d",       0xf0ff, 0x001a, 0,      0 },
        { N_F,      "sts pr, r%d",         0xf0ff, 0x002a, 0,      0 },
        { N_F,      "tas.b @r%d",          0xf0ff, 0x401b, 0,      0 },
        { N_F,      "stc.l sr, @-r%d",     0xf0ff, 0x4003, 0,      0 },
        { N_F,      "stc.l gbr, @-r%d",    0xf0ff, 0x4013, 0,      0 },
        { N_F,      "stc.l vbr, @-r%d",    0xf0ff, 0x4023, 0,      0 },
        { N_F,      "sts.l mach, @-r%d",   0xf0ff, 0x4002, 0,      0 },
        { N_F,      "sts.l macl, @-r%d",   0xf0ff, 0x4012, 0,      0 },
        { N_F,      "sts.l pr, @-r%d",     0xf0ff, 0x4022, 0,      0 },
        { M_F,      "ldc r%d, sr",         0xf0ff, 0x400e, 0,      0 },
        { M_F,      "ldc r%d, gbr",        0xf0ff, 0x401e, 0,      0 },
        { M_F,      "ldc r%d, vbr",        0xf0ff, 0x402e, 0,      0 },
        { M_F,      "lds r%d, mach",       0xf0ff, 0x400a, 0,      0 },
        { M_F,      "lds r%d, macl",       0xf0ff, 0x401a, 0,      0 },
        { M_F,      "lds r%d, pr",         0xf0ff, 0x402a, 0,      0 },
        { M_F,      "jmp @r%d",            0xf0ff, 0x402b, 0,      0 },
        { M_F,      "jsr @r%d",            0xf0ff, 0x400b, 0,      0 },
        { M_F,      "ldc.l @r%d+, sr",     0xf0ff, 0x4007, 0,      0 },
        { M_F,      "ldc.l @r%d+, gbr",    0xf0ff, 0x4017, 0,      0 },
        { M_F,      "ldc.l @r%d+, vbr",    0xf0ff, 0x4027, 0,      0 },
        { M_F,      "lds.l @r%d+, mach",   0xf0ff, 0x4006, 0,      0 },
        { M_F,      "lds.l @r%d+, macl",   0xf0ff, 0x4016, 0,      0 },
        { M_F,      "lds.l @r%d+, pr",     0xf0ff, 0x4026, 0,      0 },
        { M_F,      "braf r%d",            0xf0ff, 0x0023, 0,      0 },
        { M_F,      "bsrf r%d",            0xf0ff, 0x0003, 0,      0 },
        { NM_F,     "add r%d, r%d",        0xf00f, 0x300c, 0,      0 },
        { NM_F,     "addc r%d, r%d",       0xf00f, 0x300e, 0,      0 },
        { NM_F,     "addv r%d, r%d",       0xf00f, 0x300f, 0,      0 },
        { NM_F,     "and r%d, r%d",        0xf00f, 0x2009, 0,      0 },
        { NM_F,     "cmp/eq r%d, r%d",     0xf00f, 0x3000, 0,      0 },
        { NM_F,     "cmp/hs r%d, r%d",     0xf00f, 0x3002, 0,      0 },
        { NM_F,     "cmp/ge r%d, r%d",     0xf00f, 0x3003, 0,      0 },
        { NM_F,     "cmp/hi r%d, r%d",     0xf00f, 0x3006, 0,      0 },
        { NM_F,     "cmp/gt r%d, r%d",     0xf00f, 0x3007, 0,      0 },
        { NM_F,     "cmp/str r%d, r%d",    0xf00f, 0x200c, 0,      0 },
        { NM_F,     "div1 r%d, r%d",       0xf00f, 0x3004, 0,      0 },
        { NM_F,     "div0s r%d, r%d",      0xf00f, 0x2007, 0,      0 },
        { NM_F,     "dmuls.l r%d, r%d",    0xf00f, 0x300d, 0,      0 },
        { NM_F,     "dmulu.l r%d, r%d",    0xf00f, 0x3005, 0,      0 },
        { NM_F,     "exts.b r%d, r%d",     0xf00f, 0x600e, 0,      0 },
        { NM_F,     "exts.w r%d, r%d",     0xf00f, 0x600f, 0,      0 },
        { NM_F,     "extu.b r%d, r%d",     0xf00f, 0x600c, 0,      0 },
        { NM_F,     "extu.w r%d, r%d",     0xf00f, 0x600d, 0,      0 },
        { NM_F,     "mov r%d, r%d",        0xf00f, 0x6003, 0,      0 },
        { NM_F,     "mul.l r%d, r%d",      0xf00f, 0x0007, 0,      0 },
        { NM_F,     "muls.w r%d, r%d",     0xf00f, 0x200f, 0,      0 },
        { NM_F,     "mulu.w r%d, r%d",     0xf00f, 0x200e, 0,      0 },
        { NM_F,     "neg r%d, r%d",        0xf00f, 0x600b, 0,      0 },
        { NM_F,     "negc r%d, r%d",       0xf00f, 0x600a, 0,      0 },
        { NM_F,     "not r%d, r%d",        0xf00f, 0x6007, 0,      0 },
        { NM_F,     "or r%d, r%d",         0xf00f, 0x200b, 0,      0 },
        { NM_F,     "sub r%d, r%d",        0xf00f, 0x3008, 0,      0 },
        { NM_F,     "subc r%d, r%d",       0xf00f, 0x300a, 0,      0 },
        { NM_F,     "subv r%d, r%d",       0xf00f, 0x300b, 0,      0 },
        { NM_F,     "swap.b r%d, r%d",     0xf00f, 0x6008, 0,      0 },
        { NM_F,     "swap.w r%d, r%d",     0xf00f, 0x6009, 0,      0 },
        { NM_F,     "tst r%d, r%d",        0xf00f, 0x2008, 0,      0 },
        { NM_F,     "xor r%d, r%d",        0xf00f, 0x200a, 0,      0 },
        { NM_F,     "xtrct r%d, r%d",      0xf00f, 0x200d, 0,      0 },
        { NM_F,     "mov.b r%d, @r%d",     0xf00f, 0x2000, 0,      0 },
        { NM_F,     "mov.w r%d, @r%d",     0xf00f, 0x2001, 0,      0 },
        { NM_F,     "mov.l r%d, @r%d",     0xf00f, 0x2002, 0,      0 },
        { NM_F,     "mov.b @r%d, r%d",     0xf00f, 0x6000, 0,      0 },
        { NM_F,     "mov.w @r%d, r%d",     0xf00f, 0x6001, 0,      0 },
        { NM_F,     "mov.l @r%d, r%d",     0xf00f, 0x6002, 0,      0 },
        { NM_F,     "mac.l @r%d+, @r%d+",  0xf00f, 0x000f, 0,      0 },
        { NM_F,     "mac.w @r%d+, @r%d+",  0xf00f, 0x400f, 0,      0 },
        { NM_F,     "mov.b @r%d+, r%d",    0xf00f, 0x6004, 0,      0 },
        { NM_F,     "mov.w @r%d+, r%d",    0xf00f, 0x6005, 0,      0 },
        { NM_F,     "mov.l @r%d+, r%d",    0xf00f, 0x6006, 0,      0 },
        { NM_F,     "mov.b r%d, @-r%d",    0xf00f, 0x2004, 0,      0 },
        { NM_F,     "mov.w r%d, @-r%d",    0xf00f, 0x2005, 0,      0 },
        { NM_F,     "mov.l r%d, @-r%d",    0xf00f, 0x2006, 0,      0 },
        { NM_F,     "mov.b r%d, @(r0, r%d)", 0xf00f, 0x0004, 0,    0 },
        { NM_F,     "mov.w r%d, @(r0, r%d)", 0xf00f, 0x0005, 0,    0 },
        { NM_F,     "mov.l r%d, @(r0, r%d)", 0xf00f, 0x0006, 0,    0 },
        { NM_F,     "mov.b @(r0, r%d), r%d", 0xf00f, 0x000c, 0,    0 },
        { NM_F,     "mov.w @(r0, r%d), r%d", 0xf00f, 0x000d, 0,    0 },
        { NM_F,     "mov.l @(r0, r%d), r%d", 0xf00f, 0x000e, 0,    0 },
        { MD_F,     "mov.b @(0x%03X, r%d), r0", 0xff00, 0x8400, 0, 0 },
        { MD_F,     "mov.w @(0x%03X, r%d), r0", 0xff00, 0x8500, 0, 0 },
        { ND4_F,    "mov.b r0, @(0x%03X, r%d)", 0xff00, 0x8000, 0, 0 },
        { ND4_F,    "mov.w r0, @(0x%03X, r%d)", 0xff00, 0x8100, 0, 0 },
        { NMD_F,    "mov.l r%d, @(0x%03X, r%d)", 0xf000, 0x1000, 0,0 },
        { NMD_F,    "mov.l @(0x%03X, r%d), r%d", 0xf000, 0x5000, 0,0 },
        { D_F,      "mov.b r0, @(0x%03X, gbr)", 0xff00, 0xc000, 1, 0 },
        { D_F,      "mov.w r0, @(0x%03X, gbr)", 0xff00, 0xc100, 2, 0 },
        { D_F,      "mov.l r0, @(0x%03X, gbr)", 0xff00, 0xc200, 4, 0 },
        { D_F,      "mov.b @(0x%03X, gbr), r0", 0xff00, 0xc400, 1, 0 },
        { D_F,      "mov.w @(0x%03X, gbr), r0", 0xff00, 0xc500, 2, 0 },
        { D_F,      "mov.l @(0x%03X, gbr), r0", 0xff00, 0xc600, 4, 0 },
        { D_F,      "mova @(0x%03X, pc), r0", 0xff00, 0xc700, 4,   0 },
        { D_F,      "bf 0x%08X",           0xff00, 0x8b00, 5,      0 },
        { D_F,      "bf/s 0x%08X",         0xff00, 0x8f00, 5,      0 },
        { D_F,      "bt 0x%08X",           0xff00, 0x8900, 5,      0 },
        { D_F,      "bt/s 0x%08X",         0xff00, 0x8d00, 5,      0 },
        { D12_F,    "bra 0x%08X",          0xf000, 0xa000, 0,      0 },
        { D12_F,    "bsr 0x%08X",          0xf000, 0xb000, 0,      0 },
        { ND8_F,    "mov.w @(0x%03X, pc), r%d", 0xf000, 0x9000, 2, 0 },
        { ND8_F,    "mov.l @(0x%03X, pc), r%d", 0xf000, 0xd000, 4, 0 },
        { I_F,      "and.b #0x%02X, @(r0, gbr)", 0xff00, 0xcd00, 0,0 },
        { I_F,      "or.b #0x%02X, @(r0, gbr)",  0xff00, 0xcf00, 0,0 },
        { I_F,      "tst.b #0x%02X, @(r0, gbr)", 0xff00, 0xcc00, 0,0 },
        { I_F,      "xor.b #0x%02X, @(r0, gbr)", 0xff00, 0xce00, 0,0 },
        { I_F,      "and #0x%02X, r0",     0xff00, 0xc900, 0,      0 },
        { I_F,      "cmp/eq #0x%02X, r0",  0xff00, 0x8800, 0,      0 },
        { I_F,      "or #0x%02X, r0",      0xff00, 0xcb00, 0,      0 },
        { I_F,      "tst #0x%02X, r0",     0xff00, 0xc800, 0,      0 },
        { I_F,      "xor #0x%02X, r0",     0xff00, 0xca00, 0,      0 },
        { I_F,      "trapa #0x%X",         0xff00, 0xc300, 0,      0 },
        { NI_F,     "add #0x%02X, r%d",    0xf000, 0x7000, 0,      0 },
        { NI_F,     "mov #0x%02X, r%d",    0xf000, 0xe000, 0,      0 },
        { 0,        NULL,                  0,      0,      0,      0 }
};

void ResetCPU(tagSH2 *sh2)
{
	int i;

	memset((void*)sh2->GenReg, 0, sizeof(uint32)*16);
	memset((void*)sh2->CtrlReg, 0, sizeof(uint32)*3);
	memset((void*)sh2->SysReg, 0, sizeof(uint32)*4);

	// SMPC Stuff
	for (i=1; i<0x80*2; i+=2) 
		*(unsigned short*)&SMPC[i] = 0xff;

	SMPC[0x5f] = 0x06;

	sh2->CtrlReg[0] = 0xf;
	sh2->GenReg[15] = 0x06002000;
}

uint32	temp, DelayWait;
uint64	mult;

void doInst(tagSH2 *sh2, unsigned short opcode)
{
	//unsigned long temp;
	switch((opcode >> 12) & 0xf)
	{
		case (0x0):
			switch (opcode & 0xf) {
				case (0x2):
					if (source == 0) //STC SR,Rn
						sh2->GenReg[dest] = SR;
					else if (source == 1) //STC GBR,Rn
						sh2->GenReg[dest] = GBR;
					else if (source == 2) //STC VBR,Rn
						sh2->GenReg[dest] = VBR;
					break;
				case (0x3):
					if ((opcode >> 4) & 0xf == 0) //BSRF
					{
						PR = PC + 4;
						temp = PC + sh2->GenReg[dest];
						DelayBranch(temp);
					}
					if ((opcode >> 4) & 0xf == 2) //BRAF
					{
						temp = PC += sh2->GenReg[dest];
						DelayBranch(temp)
					}
					break;
				case (0x4): //MOV.B  Rm,@(R0,Rn)
					memSetByte(sh2->GenReg[0]+sh2->GenReg[dest], (uint8)sh2->GenReg[source]);
					break;
				case (0x5): //MOV.W  Rm,@(R0,Rn)
					memSetWord(sh2->GenReg[0]+sh2->GenReg[dest], (uint16)sh2->GenReg[source]);
					break;
				case (0x6): //MOV.L  Rm,@(R0,Rn)
					memSetLong(sh2->GenReg[0]+sh2->GenReg[dest], sh2->GenReg[source]);
					break;
				case (0x7): //MUL.L Rm,Rn
					MACL = (uint32)(sh2->GenReg[source] * sh2->GenReg[dest]);
					break;
				case (0x8):
					if (opcode == 0x8) //CLRT
						SR &= 0xfffffffe;
					else if (opcode == 0x18) //SETT
						setTBit(1);
					else if (opcode == 0x28) //CLRMAC
						MACH = MACL = 0;
					break;
				case (0x9):
					if (opcode == 0x9) //NOP
					{}
					else if (opcode == 0x19) //DIV0U
						SR &= 0xfcfe;
					else if (opcode & 0xff == 0x29) //MOVT
						sh2->GenReg[dest] = SR & 1;
					break;
				case (0xa):
					if (((opcode >> 4) & 0xf) == 0) //STS MACH,Rn
						sh2->GenReg[dest] = MACH;
					else if (((opcode >> 4) & 0xf) == 1) //STS MACL,Rn
						sh2->GenReg[dest] = MACL;
					else if (((opcode >> 4) & 0xf) == 2) //STS PR,Rn
						sh2->GenReg[dest] = PR;
					break;
				case (0xb):
					if (opcode == 0xb) //RTS
					{
						temp = PR;
						DelayBranch(temp);
					}
					else if (opcode == 0x1b) //SLEEP
					{
						ConsoleMsg(MSG_WARN, "SLEEP Opcode Interception");
					}
					else if (opcode == 0x2b) //RFE
					{
						temp = memGetLong(sh2->GenReg[15]) + 4;
						sh2->GenReg[15] += 4;
						SR = memGetLong(sh2->GenReg[15]) & 0x3f3;
						sh2->GenReg[15] += 4;
						DelayBranch(temp);
					}
					break;
				case (0xc): //MOV.B @(R0,Rm),Rn
					sh2->GenReg[dest] = (signed long)memGetByte(sh2->GenReg[0]+sh2->GenReg[source]);
					break;
				case (0xd): //MOV.W @(R0,Rm),Rn
					sh2->GenReg[dest] = (signed long)memGetWord(sh2->GenReg[0]+sh2->GenReg[source]);
					break;
				case (0xe): //MOV.L @(R0,Rm),Rn
					sh2->GenReg[dest] = memGetLong(sh2->GenReg[0]+sh2->GenReg[source]);
					break;
				case (0xf): //MAC.L @Rm+,@Rn+  **test
					mult = (uint64)((signed long)memGetLong(sh2->GenReg[source])*(signed long)memGetLong(sh2->GenReg[dest]));
					sh2->GenReg[source] += 4;
					sh2->GenReg[dest] += 4;
					mult += ((uint64)(MACH << 32) | (uint64)(MACL));
					MACL = (uint32)(mult & 0xffffffff);
					MACH = (uint32)((mult >> 32) & 0xffffffff);
					if (SR&1)
						MACH &= 0xffff;
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%4x", PC, opcode);
			}
			break;
		case (0x1):
			memSetLong(sh2->GenReg[dest]+(offset1<<2), sh2->GenReg[source]);
			break;
		case (0x2):
			switch (opcode & 0xf) {
				case (0x0): //MOV.B
					memSetByte(sh2->GenReg[dest], (uint8)sh2->GenReg[source]);
					break;
				case (0x1): //MOV.W
					memSetWord(sh2->GenReg[dest], (uint16)sh2->GenReg[source]);
					break;
				case (0x2): //MOV.L
					memSetLong(sh2->GenReg[dest], sh2->GenReg[source]);
					break;
				case (0x4): //MOV.B
					memSetByte(--sh2->GenReg[dest], (uint8)sh2->GenReg[source]);
					break;
				case (0x5): //MOV.W
					sh2->GenReg[dest] -= 2;
					memSetWord(sh2->GenReg[dest], (uint16)sh2->GenReg[source]);
					break;
				case (0x6): //MOV.L
					sh2->GenReg[dest] -= 4;
					memSetLong(sh2->GenReg[dest], sh2->GenReg[source]);
					break;
				case (0x7): //DIV0S Rm,Rn
					if ((sh2->GenReg[source]&0x80000000)^(sh2->GenReg[dest]&0x80000000))
						setTBit(0);
					else
						setTBit(1);
					break;
				case (0x8): //TST Rm,Rn
					if ((sh2->GenReg[dest] & sh2->GenReg[source]) == 0)
						setTBit(1);
					else
						setTBit(0);
					break;
				case (0x9): //AND Rm,Rn
					sh2->GenReg[dest] &= sh2->GenReg[source];
					break;
				case (0xa): //XOR Rm,Rn
					sh2->GenReg[dest] ^= sh2->GenReg[source];
					break;
				case (0xb): //OR Rm,Rn
					sh2->GenReg[dest] |= sh2->GenReg[source];
					break;
				case (0xc): //CMP/STR Rm,Rn
					temp = sh2->GenReg[dest]^sh2->GenReg[source];
					if ((temp>>12)&0xff || (temp>>8)&0xff || (temp>>4)&0xff || temp&0xff)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0xd): //XTRCT Rm,Rn
					sh2->GenReg[dest] = ((sh2->GenReg[source]<<16)&0xffff0000) | ((sh2->GenReg[dest]>>16)&0xffff);
					break;
				case (0xe): //MULU Rm,Rn
					MACL=((uint32)(uint16)sh2->GenReg[dest]*(uint32)(uint16)sh2->GenReg[source]);
					break;
				case (0xf): //MULS Rm,Rn
					MACL=((signed long)(short)sh2->GenReg[dest]*(signed long)(short)sh2->GenReg[source]);
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%4x", PC, opcode);
			}
			break;
		case (0x3):
			switch (opcode & 0xf) {
				case (0x0): //CMP/EQ Rm,Rn
					if (sh2->GenReg[source] == sh2->GenReg[dest])
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x2): //CMP/HS Rm,Rn
					if (sh2->GenReg[dest] >= sh2->GenReg[source])
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x3): //CMP/GE Rm,Rn
					if ((signed)sh2->GenReg[dest] >= (signed)sh2->GenReg[source])
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x4): //DIV1 Rm, Rn
					// *(unsigned long*)&mult is used to generate 32-bit code
					temp = (0x80000000 & sh2->GenReg[dest]) != 0;
					sh2->GenReg[dest] = ((sh2->GenReg[dest]<<1) | (SR&1));

					switch((SR&0x100 != 0)) {
						case 0:
							switch((SR&0x200 != 0)) {
								case 0:
									*(unsigned long*)&mult = sh2->GenReg[dest];
									sh2->GenReg[dest] -= sh2->GenReg[source];
									setQBit((sh2->GenReg[dest] > (uint32)mult) ^ temp);
									break;
								case 1:
									*(unsigned long*)&mult = sh2->GenReg[dest];
									sh2->GenReg[dest] += sh2->GenReg[source];
									setQBit((sh2->GenReg[dest] >= (uint32)mult) ^ temp);
									break;
							}
							break;
						case 1:
							switch((SR&0x200 != 0)) {
								case 0:
									*(unsigned long*)&mult = sh2->GenReg[dest];
									sh2->GenReg[dest] += sh2->GenReg[source];
									setQBit((sh2->GenReg[dest] < (uint32)mult) ^ temp);
									break;
								case 1:
									*(unsigned long*)&mult = sh2->GenReg[dest];
									sh2->GenReg[dest] -= sh2->GenReg[source];
									setQBit((sh2->GenReg[dest] <= (uint32)mult) ^ temp);
									break;
							}
							break;
					}
					setTBit((SR&0x100)^(SR&0x200));

					break;
				case (0x5): //DMULU.L Rm, Rn
					mult = ((uint64)sh2->GenReg[source] * (uint64)sh2->GenReg[dest]);
					MACL = (uint32)mult;
					MACH = (uint32)(mult >> 32);
					break;
				case (0x6): //CMP/HI Rm,Rn
					if (sh2->GenReg[dest] > sh2->GenReg[source])
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x7): //CMP/GT Rm,Rn
					if ((signed)sh2->GenReg[dest] > (signed)sh2->GenReg[source])
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x8): //SUB Rm,Rn
					sh2->GenReg[dest] -= sh2->GenReg[source];
					break;
				case (0xa): //SUBC Rm,Rn
					temp = sh2->GenReg[dest] - sh2->GenReg[source];
					if ((sh2->GenReg[dest] < temp) || (temp < temp - (SR&1)))
					{
						sh2->GenReg[dest] = temp - SR&1;
						setTBit(1);
					} else 
					{
						sh2->GenReg[dest] = temp - SR&1;
						setTBit(0);
					}
					break;
				case (0xb): //SUBV Rm,Rn **test
					temp = sh2->GenReg[dest] - sh2->GenReg[source];
					if (((signed long)(sh2->GenReg[dest] < 0) ^ (signed long)(sh2->GenReg[source] < 0)) && (temp < 0))
						setTBit(1);
					else
						setTBit(0);
					sh2->GenReg[dest] = temp;
					break;
				case (0xc): //ADD Rm,Rn
					sh2->GenReg[dest] += sh2->GenReg[source];
					break;
				case (0xd): //DMULS.L Rm, Rn
					mult = ((sint64)sh2->GenReg[source] * (sint64)sh2->GenReg[dest]);
					MACL = (uint32)mult;
					MACH = (uint32)(mult >> 32);
					break;
				case (0xe): //ADDC Rm,Rn **test
					temp = (uint32)(sh2->GenReg[dest] + sh2->GenReg[source]);
					if ((temp+(SR&1) < temp) || (temp < sh2->GenReg[dest]))
					{
						sh2->GenReg[dest] = temp + (SR&1);
						setTBit(1);
					} else {
						sh2->GenReg[dest] = temp + (SR&1);
						SR &= 0xfffffffe;
					}
					break;
				case (0xf): //ADDV Rm,Rn
					temp = sh2->GenReg[source] + sh2->GenReg[dest];
					if (!(((signed long)(sh2->GenReg[source] < 0) ^ (signed long)(sh2->GenReg[dest] < 0))) && (temp < 0))
						setTBit(1);
					else
						setTBit(0);
					sh2->GenReg[dest] = temp;
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%4x", PC, opcode);
			}
			break;
		case (0x4):
			switch (opcode & 0xff) {
				case (0x0):
				case (0x20): //SHLL Rn (0x0) <=> SHAL Rn (0x20)
					setTBit((sh2->GenReg[dest] >> 31) & 1);
					sh2->GenReg[dest] <<= 1;
					break;
				case (0x1): //SHLR Rn
					setTBit(sh2->GenReg[dest] & 1);
					sh2->GenReg[dest] >>= 1;
					break;
				case (0x2): //STS.L MACH,@-Rn
					sh2->GenReg[dest] -= 4;
					memSetLong(sh2->GenReg[dest], MACH);
					break;
				case (0x3): 
					sh2->GenReg[dest] -= 4;
					temp = (opcode >> 4) & 0xff;
					if (temp == 0) // STC.L SR,@-Rn
						memSetLong(sh2->GenReg[dest], SR);
					else if (temp == 1) //STC GBR,@-Rn
						memSetLong(sh2->GenReg[dest], GBR);
					else if (temp == 2) //STC VBR,@-Rn
						memSetLong(sh2->GenReg[dest], VBR);
					break;
				case (0x4): //ROTL Rn
					temp = ((sh2->GenReg[dest] >> 31) & 1);
					sh2->GenReg[dest] = (sh2->GenReg[dest] << 1) | (SR&1);
					setTBit(temp);
					break;
				case (0x5): //ROTR Rn
					temp = sh2->GenReg[dest] & 1;
					sh2->GenReg[dest] = (sh2->GenReg[dest] >> 1) | (uint32)((SR&1)<<31);
					setTBit(temp);
					break;
				case (0x6): //LDS.L @Rm+,MACH
					MACH = memGetLong(sh2->GenReg[dest]);
					sh2->GenReg[dest] += 4;
					break;
				case (0x7): //LDC.L @Rm+,SR
					SR = memGetLong(sh2->GenReg[dest]) & 0x3f3;
					sh2->GenReg[dest] += 4;
					break;
				case (0x8): //SHLL2 Rn
					sh2->GenReg[dest] <<= 2;
					break;
				case (0x9): //SHLR2 Rn
					sh2->GenReg[dest] >>= 2;
					break;
				case (0xa): //LDS Rm,MACH
					MACH = sh2->GenReg[dest];
					break;
				case (0xb): //JSR @Rm
					PR = PC + 4;
					temp = sh2->GenReg[dest];
					DelayBranch(temp);
					break;
				case (0xe): //LDC Rm,SR
					SR = sh2->GenReg[dest] & 0x3f3;
					break;
				case (0xf): //MAC.W @Rm+,@Rn+
					temp = ((signed long)memGetWord(sh2->GenReg[dest])*(signed long)memGetWord(sh2->GenReg[source]));
					sh2->GenReg[dest] += 4;
					sh2->GenReg[source] += 4;

					if (SR&2)
					{
						MACL = (uint32)((signed)MACL + (signed)temp);
						if (MACL > 0x80000000)
						{
							MACH = 1;
							MACL &= 0x8fffffff;
						}
					}
					else
					{
						mult = ((uint64)(MACH<<32))+MACL+temp;
						MACL = (uint32)mult;
						MACH = (uint32)(mult>>32);
					}
					break;
				case (0x10): //DT Rn
					if (--sh2->GenReg[dest] == 0)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x11): //CMP/PZ Rn
					if ((signed)sh2->GenReg[dest] >= 0)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x12): //STS.L MACL,@-Rn
					sh2->GenReg[dest] -= 4;
					memSetLong(sh2->GenReg[dest], MACL);
					break;
				case (0x15): //CMP/PL Rn
					if ((signed)sh2->GenReg[dest] > 0)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x16): //LDS.L @Rm+,MACL
					MACL = memGetLong(sh2->GenReg[dest]);
					sh2->GenReg[dest] += 4;
					break;
				case (0x17): //LDC.L @Rm+,GBR
					GBR = memGetLong(sh2->GenReg[dest]);
					sh2->GenReg[dest] += 4;
					break;
				case (0x18): //SHLL8 Rn
					sh2->GenReg[dest] <<= 8;
					break;
				case (0x19): //SHLR8 Rn
					sh2->GenReg[dest] >>= 8;
					break;
				case (0x1a): //LDS Rm,MACL
					MACL = sh2->GenReg[dest];
					break;
				case (0x1b): //TAS.B @Rn
					temp = (uint32)memGetByte(sh2->GenReg[dest]);
					if (temp == 0) setTBit(1);
					else setTBit(0);
					temp |= 0x80;
					memSetByte(sh2->GenReg[dest], (uint8)temp);
					break;
				case (0x1e): //LDC Rm,GBR
					GBR = sh2->GenReg[dest];
					break;
				case (0x21): //SHAR Rn
					temp = sh2->GenReg[dest] & 0x80000000;
					setTBit(sh2->GenReg[dest] & 1);
					sh2->GenReg[dest] = ((sh2->GenReg[dest] >> 1) & 0x7fffffff) | temp;
					break;
				case (0x22): //STS.L PR,@-Rn
					sh2->GenReg[dest] -= 4;
					memSetLong(sh2->GenReg[dest], PR);
					break;
				case (0x24): //ROTCL Rn
					temp = (sh2->GenReg[dest] >> 31) & 1;
					sh2->GenReg[dest] = (sh2->GenReg[dest] << 1) | (SR&1);
					setTBit(temp);
					break;
				case (0x25): //ROTCR Rn
					temp = sh2->GenReg[dest] & 1;
					sh2->GenReg[dest] = (sh2->GenReg[dest] >> 1) | (uint32)((SR&1) << 31);
					setTBit(temp);
					break;
				case (0x26): //LDS.L @Rm+,PR
					PR = memGetLong(sh2->GenReg[dest]);
					sh2->GenReg[dest] += 4;
					break;
				case (0x27): //LDC.L @Rm+,VBR
					VBR = memGetLong(sh2->GenReg[dest]);
					sh2->GenReg[dest] += 4;
					break;
				case (0x28): //SHLL16 Rn
					sh2->GenReg[dest] <<= 16;
					break;
				case (0x29): //SHLR16 Rn
					sh2->GenReg[dest] >>= 16;
					break;
				case (0x2a): //LDS Rm,PR
					PR = sh2->GenReg[dest];
					break;
				case (0x2b): //JMP @Rm
					temp = sh2->GenReg[dest];
					DelayBranch(temp);
					break;
				case (0x2e): //LDC Rm,VBR
					VBR = sh2->GenReg[dest];
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%4x", PC, opcode);
			}
			break;
		case (0x5): //MOV.L @(disp,Rm),Rn
			sh2->GenReg[dest] = memGetLong(sh2->GenReg[source] + (offset1<<2));
			break;
		case (0x6): //MOV
			switch (opcode & 0xf) {
				case (0x0): //MOV.B @Rm,Rn
					sh2->GenReg[dest] = (signed char)memGetByte(sh2->GenReg[source]);
					break;
				case (0x1): //MOV.W @Rm,Rn
					sh2->GenReg[dest] = (signed short)memGetWord(sh2->GenReg[source]);
					break;
				case (0x2): //MOV.L
					sh2->GenReg[dest] = memGetLong(sh2->GenReg[source]);
					break;
				case (0x3): //MOV
					sh2->GenReg[dest] = sh2->GenReg[source];
					break;
				case (0x4): //MOV.B
					sh2->GenReg[dest] = (signed char)memGetByte(sh2->GenReg[source]++);
					break;
				case (0x5): //MOV.W
					sh2->GenReg[dest] = (signed short)memGetWord(sh2->GenReg[source]);
					sh2->GenReg[source] += 2;
					break;
				case (0x6): //MOV.L @Rm+,Rn
					sh2->GenReg[dest] = (signed long)memGetLong(sh2->GenReg[source]);
					sh2->GenReg[source] += 4;
					break;
				case (0x7): //NOT Rm,Rn
					sh2->GenReg[dest] = ~sh2->GenReg[source];
					break;
				case (0x8): //SWAP.B
					sh2->GenReg[dest] = (sh2->GenReg[dest] & 0xffff0000)|(uint16)(((uint16)sh2->GenReg[source]>>8)|((uint16)sh2->GenReg[source]<<8));
					break;
				case (0x9): //SWAP.W
					sh2->GenReg[dest] = (sh2->GenReg[source] >> 10) | (sh2->GenReg[source] << 10);
					break;
				case (0xa): //NEGC Rm,Rn
					temp = (signed long)(0-sh2->GenReg[source]);
					sh2->GenReg[dest] = (signed long)(temp) - (SR&1);
					if (0 < (signed)temp)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					if (temp < sh2->GenReg[dest]) setTBit(1);
					break;
				case (0xb): //NEG Rm,Rn
					sh2->GenReg[dest] = (signed long)(0-sh2->GenReg[source]);
					break;
				case (0xc): //EXTU.B Rm,Rn
					sh2->GenReg[dest] = sh2->GenReg[source] & 0xff;
					break;
				case (0xd): //EXTU.W Rm,Rn
					sh2->GenReg[dest] = sh2->GenReg[source] & 0xffff;
					break;
				case (0xe): //EXTS.B Rm,Rn
					sh2->GenReg[dest] = (signed long)((char)sh2->GenReg[source]);
					break;
				case (0xf): //EXTS.W Rm,Rn
					sh2->GenReg[dest] = (signed long)((short)sh2->GenReg[source]);
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%4x", PC, opcode);
			}
			break;
		case (0x7):	//ADDI #imm,Rn
			sh2->GenReg[dest] += (signed long)immediate;
			break;
		case (0x8): //MOV
			switch ((opcode >> 8) & 0xf) {
				case (0x0): //MOV.B
					memSetByte(sh2->GenReg[source]+offset1, (uint8)sh2->GenReg[0]);
					break;
				case (0x1): //MOV.W
					memSetWord(sh2->GenReg[source]+(offset1<<1), (uint16)sh2->GenReg[0]);
					break;
				case (0x4): //MOV.B
					sh2->GenReg[0] = (signed long)memGetByte(sh2->GenReg[source]+offset1);
					break;
				case (0x5): //MOV.W
					sh2->GenReg[0] = (signed long)memGetWord(sh2->GenReg[source]+(offset1<<1));
					break;
				case (0x8): //CMP_EQ #imm,R0
					if ((signed)sh2->GenReg[0] == (signed long)immediate)
						setTBit(1);
					else
						SR &= 0xfffffffe;
					break;
				case (0x9): //BT
					if (SR&1)
						PC += (signed char)((immediate << 1) + 4) - 2;
					break;
				case (0xb): //BF
					if (!(SR&1))
						PC += (signed char)((immediate << 1) + 4) - 2;
					break;
				case (0xd): //BT/S
					if (SR&1)
					{
						temp = PC + (signed char)((immediate << 1) + 4);
						DelayBranch(temp);
					}
					break;
				case (0xf): //BF/S
					if (!(SR&1))
					{
						temp = PC + (signed char)((immediate << 1) + 4);
						DelayBranch(temp);
					}
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%8x", PC, opcode);
			}
			break;
		case (0x9): //MOV.W
			sh2->GenReg[dest] = (signed long)memGetWord(PC+(offset2<<1)+4);
			break;
		case (0xa): //BRA
			temp = (offset3 << 1);
			if (temp & 0x1000)
				temp |= 0xfffff000;
			temp = PC + (signed long)temp + 4;
			DelayBranch(temp);
			break;
		case (0xb): //BSR
			PR = PC + 4;
			temp = (offset3 << 1);
			if (temp & 0x1000)
				temp |= 0xfffff000;
			temp = PC + (signed long)temp + 4;
			DelayBranch(temp);
			break;
		case (0xc):
			switch ((opcode >> 8) & 0xf) {
				case (0x0): //MOV.B  R0,@(disp,GBR)
					memSetByte(offset2+GBR, (uint8)sh2->GenReg[0]);
					break;
				case (0x1): //MOV.W  R0,@(disp,GBR)
					memSetWord((offset2<<1)+GBR, (uint16)sh2->GenReg[0]);
					break;
				case (0x2): //MOV.L  R0,@(disp,GBR)
					memSetLong((offset2<<2)+GBR, sh2->GenReg[0]);
					break;
				case (0x3): //TRAPA #imm
					sh2->GenReg[15] -= 4;
					memSetLong(sh2->GenReg[15], SR);
					sh2->GenReg[15] -= 4;
					memSetLong(sh2->GenReg[15], PC-2);
					PC = memGetLong(VBR + ((uint32)immediate << 2)) + 4 - 2;
					break;
				case (0x4): //MOV.B  @(disp,GBR),R0
					sh2->GenReg[0] = (signed long)memGetByte(offset2+GBR);
					break;
				case (0x5): //MOV.W  @(disp,GBR),R0
					sh2->GenReg[0] = (signed long)memGetWord((offset2<<1)+GBR);
					break;
				case (0x6): //MOV.L  @(disp,GBR),R0
					sh2->GenReg[0] = memGetLong((offset2<<2)+GBR);
					break;
				case (0x7): //MOVA
					sh2->GenReg[0] = (PC&0xfffffffc)+(offset2<<2)+4;
					break;
				case (0x8): //TST #imm,R0
					if (((uint8)sh2->GenReg[0] & offset2) == 0)
						setTBit(1);
					else
						setTBit(0);
					break;
				case (0x9): //AND #imm,R0
					sh2->GenReg[0] &= (uint32)immediate;
					break;
				case (0xa): //XOR #imm,R0
					sh2->GenReg[0] ^= (uint32)immediate;
					break;
				case (0xb): //ORI #imm,R0
					sh2->GenReg[0] |= (uint32)immediate;
					break;
				case (0xc): //TST.B #imm,@(R0,GBR)
					if (((uint32)memGetByte(GBR + sh2->GenReg[0]) & (uint32)immediate) == 0)
						setTBit(1);
					else
						setTBit(0);
					break;
				case (0xd): //AND.B #imm,@(R0,GBR)
					memSetByte(GBR + sh2->GenReg[0], (uint8)((uint32)memGetByte(GBR + sh2->GenReg[0]) & (uint32)immediate));
					break;
				case (0xe): //XOR.B #imm,@(R0,GBR)
					memSetByte(GBR + sh2->GenReg[0], (uint8)((uint32)memGetByte(GBR + sh2->GenReg[0]) ^ (uint32)immediate));
					break;
				case (0xf): //OR.B #imm,@(R0,GBR)
					memSetByte(GBR + sh2->GenReg[0], (uint8)((uint32)memGetByte(GBR + sh2->GenReg[0]) | (0x000000FF & (uint32)immediate)));
					break;
				default:
					ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: 0x%8x", PC, opcode);
			}
			break;
		case (0xd): //MOV.L
			sh2->GenReg[dest] = memGetLong((PC&0xfffffffc)+(offset2<<2)+4);
			break;
		case (0xe): // MOV
			sh2->GenReg[dest] = (long)immediate;
			break;
		default:
			ConsoleMsg(MSG_WARN, "Unhandled Opcode at 0x%8x: (0x%4x)", PC, opcode);
			break;
	}
}